# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 Esteban J. G. Gabancho.
#
# Flask-SSO-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Views tests."""

import pytest
from flask import url_for
from flask_security import url_for_security
from mock import patch


def test_wrong_idp(appctx, base_client):
    """Test wrong Identity Provider name."""
    client = base_client
    login_url = url_for("sso_saml.sso", idp="wrong-idp")
    res = client.get(login_url)
    assert res.status_code == 404


def test_login_error(appctx, base_client):
    """Test login error."""
    client = base_client
    login_url = url_for("sso_saml.sso", idp="test-idp")
    with patch("invenio_saml.utils.SAMLAuth.login") as mock_login:
        mock_login.return_value = None
        res = client.get(login_url)
        assert res.status_code == 401


def test_login(appctx, base_client):
    """Test SSO requests."""
    client = base_client
    login_url = url_for("sso_saml.sso", idp="test-idp")
    res = client.get(login_url)
    assert res.status_code == 302
    assert res.location.startswith("https://test-ipd.com/sso?")

    login_url = url_for("sso_saml.sso", idp="test-idp", next="/next_url")
    res = client.get(login_url)
    assert res.status_code == 302
    assert res.location.startswith("https://test-ipd.com/sso?")
    assert "RelayState=%2Fnext_url" in res.location


@pytest.mark.freeze_time("2019-04-19T13:35:47Z")
def test_acs(appctx, base_client, sso_response):
    """Test ACS request."""
    client = base_client
    acs_url = url_for("sso_saml.acs", idp="test-idp")

    res = client.post(acs_url, data=dict(SAMLResponse=""))
    assert res.status_code == 400

    res = client.post(acs_url, data=dict(SAMLResponse=sso_response))
    assert res.status_code == 401

    with patch(
        "onelogin.saml2.auth.OneLogin_Saml2_Response.is_valid"
    ) as mock_is_valid, patch(
        "invenio_saml.utils.SAMLAuth.is_authenticated"
    ) as mock_is_authenticated:
        mock_is_valid.return_value = True
        mock_is_authenticated.return_value = False
        res = client.post(acs_url, data=dict(SAMLResponse=sso_response))
        assert res.status_code == 403

    with patch("onelogin.saml2.auth.OneLogin_Saml2_Response.is_valid") as mock_is_valid:
        mock_is_valid.return_value = True
        res = client.post(acs_url, data=dict(SAMLResponse=sso_response))
        assert res.status_code == 302


def test_logout(appctx, base_client):
    """Test SLO requests."""
    client = base_client
    logout_url = url_for("sso_saml.slo", idp="test-idp")
    with client.session_transaction() as sess:
        sess["SSO::SAML::NameId"] = "ID"
        sess["SSO::SAML::SessionIndex"] = "INDEX"
    res = client.get(logout_url)
    assert res.status_code == 302
    assert res.location.startswith("https://test-ipd.com/slo?")


def test_logout_error(appctx, base_client):
    """Test logout error."""
    client = base_client
    login_url = url_for("sso_saml.slo", idp="test-idp")
    with patch("invenio_saml.utils.SAMLAuth.logout") as mock_logout:
        mock_logout.return_value = None
        res = client.get(login_url)
        assert res.status_code == 401


def test_sls(appctx, base_client, slo_query_string):
    """Test SLS request."""
    client = base_client
    sls_url = url_for("sso_saml.sls", idp="test-idp")
    res = client.get(sls_url, query_string=slo_query_string)
    assert res.status_code == 302

    with patch("invenio_saml.utils.SAMLAuth.get_errors") as mock_get_erros, patch(
        "invenio_saml.utils.SAMLAuth.get_last_error_reason"
    ) as mock_get_reason:
        mock_get_erros.return_value = ["bad error"]
        mock_get_reason.return_value = "Test reason"
        res = client.get(sls_url, query_string=slo_query_string)
        assert res.status_code == 401
        assert res.json == ["bad error", "Test reason"]


@pytest.mark.freeze_time("2019-04-18")
def test_metadata(appctx, base_client, metadata_response):
    """Test metadata request."""
    client = base_client
    metadata_url = url_for("sso_saml.metadata", idp="test-idp")
    res = client.get(metadata_url)
    assert res.status_code == 200
    assert res.data == metadata_response

    with patch(
        "onelogin.saml2.settings.OneLogin_Saml2_Settings.validate_metadata"
    ) as mock_validate_metadata, patch(
        "invenio_saml.utils.SAMLAuth.get_last_error_reason"
    ) as mock_get_reason:
        mock_validate_metadata.return_value = ["bad error", "worst error"]
        mock_get_reason.return_value = "Test reason"
        res = client.get(metadata_url)
        assert res.status_code == 401
        assert res.json == ["bad error", "worst error", "Test reason"]


def test_login_template(appctx, base_client):
    """Test the SAML login template at least loads."""
    client = base_client
    res = client.get(url_for_security("login"))
    assert res.status_code == 200
    assert "Sign in with SAML" in res.text



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019, 2022 Esteban J. Garcia Gabancho.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import base64
from urllib.parse import urlencode

import importlib_resources as resources
import pytest
from flask_webpackext.manifest import (
    JinjaManifest,
    JinjaManifestEntry,
    JinjaManifestLoader,
)
from invenio_app.factory import create_app as create_invenio_app
from onelogin.saml2.utils import OneLogin_Saml2_Utils as saml_utils


#
# Mock the webpack manifest to avoid having to compile the full assets.
#
class MockJinjaManifest(JinjaManifest):
    """Mock manifest."""

    def __getitem__(self, key):
        """Get a manifest entry."""
        return JinjaManifestEntry(key, [key])

    def __getattr__(self, name):
        """Get a manifest entry."""
        return JinjaManifestEntry(name, [name])


class MockManifestLoader(JinjaManifestLoader):
    """Manifest loader creating a mocked manifest."""

    def load(self, filepath):
        """Load the manifest."""
        return MockJinjaManifest()


@pytest.fixture(scope="module")
def app_config(app_config):
    """Customize application configuration."""
    app_config["APP_ALLOWED_HOSTS"] = [
        "localhost",
        "example.com",
        "tests.com:5000",
    ]
    # Flask-Security has ha bug in the latest release,
    # already fixed on develop branch
    app_config["SECURITY_EMAIL_SENDER"] = "no-reply@localhost"
    app_config["WEBPACKEXT_MANIFEST_LOADER"] = MockManifestLoader
    app_config["SERVER_NAME"] = "localhost"
    app_config["SSO_SAML_DEFAULT_LOGIN_HANDLER"] = lambda auth, next_url: next_url
    app_config["SSO_SAML_IDPS"] = {
        "test-idp": {
            "settings": {
                "idp": {
                    "entityId": "https://test-idp.com",
                    "singleSignOnService": {
                        "url": "https://test-ipd.com/sso",
                    },
                    "singleLogoutService": {
                        "url": "https://test-ipd.com/slo",
                    },
                    "x509cert": "cert",
                }
            },
            "acs_handler": lambda auth, next_url: next_url,
        },
        "idp-file": {
            "settings_file_path": str(resources.files(__name__) / "data" / "idp.xml"),
            "sp_cert_file": str(resources.files(__name__) / "data" / "cert.crt"),
            "sp_key_file": str(resources.files(__name__) / "data" / "cert.key"),
        },
        "idp-url": {"settings_url": "https://test-idp.com/settings"},
    }
    # Add template
    app_config["OAUTHCLIENT_LOGIN_USER_TEMPLATE"] = "invenio_saml/login_user.html"
    app_config["THEME_FRONTPAGE"] = False
    return app_config


@pytest.fixture(scope="module")
def create_app(instance_path):
    """Application factory fixture."""
    return create_invenio_app


@pytest.fixture
def users(appctx, db):
    """Example users."""
    datastore = appctx.extensions["security"].datastore
    datastore.create_user(email="federico@example.com", password="tester", active=True)
    datastore.commit()


@pytest.fixture(scope="module")
def sso_response():
    """Mock SSO response from Identity Provider."""
    with (resources.files(__name__) / "data" / "sso_response.xml").open("rb") as f:
        return base64.b64encode(f.read())


@pytest.fixture(scope="module")
def slo_query_string():
    """Mock SLO response from Identity Provider."""
    with (resources.files(__name__) / "data" / "slo_response.xml").open() as f:
        slo_response = saml_utils.deflate_and_base64_encode(f.read())
    return urlencode(dict(SAMLResponse=slo_response))


@pytest.fixture(scope="module")
def metadata_response():
    """Metadata response."""
    with (resources.files(__name__) / "data" / "metadata.xml").open("rb") as f:
        return f.read()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2024 Esteban J. Garcia Gabancho.
# Copyright (C) 2021-2022 Graz University of Technology.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Test handlers."""

from datetime import datetime, timezone

import pytest
from flask import current_app
from flask_security import current_user, login_user
from invenio_accounts.models import User
from invenio_accounts.proxies import current_datastore
from invenio_oauthclient.models import UserIdentity
from mock import Mock, patch
from werkzeug.exceptions import Unauthorized

from invenio_saml.handlers import (
    acs_handler_factory,
    default_account_setup,
    default_sls_handler,
)


def test_default_account_setup(users):
    """Test default user account setup."""
    user = User.query.filter_by(email="federico@example.com").one()
    account_info = dict(
        external_id="123456",
        external_method="external",
        other="foo",
    )

    default_account_setup(user, account_info)

    identities = UserIdentity.query.filter_by(id_user=user.id).all()
    assert len(identities) == 1
    assert identities[0].user == user

    # If we do it again there is still one row only
    default_account_setup(user, account_info)
    identities = UserIdentity.query.filter_by(id_user=user.id).all()
    assert len(identities) == 1
    assert identities[0].user == user


def test_default_sls_handler(appctx, users):
    """Test default SLS handler."""
    with appctx.test_request_context():
        user = User.query.filter_by(email="federico@example.com").one()
        login_user(user)
        assert current_user.is_authenticated
        # call the SLS handler
        next_url = default_sls_handler(None, "/foo")
        assert not current_user.is_authenticated
        assert next_url == "/foo"


def test_acs_handler_factory(appctx, db):
    """Test ACS handler factory."""
    appctx.config["SSO_SAML_IDPS"] = {
        "test": {
            "mappings": {
                "email": "email",
                "name": "name",
                "surname": "surname",
                "external_id": "external_id",
            },
        }
    }
    attrs = dict(
        email=["federico@example.com"],
        name=["federico"],
        surname=["Fernandez"],
        external_id=["12345679abcdf"],
    )

    acs_handler = acs_handler_factory("test")

    with appctx.test_request_context(), patch(
        "invenio_saml.utils.SAMLAuth"
    ) as mock_saml_auth:
        mock_saml_auth.get_attributes.return_value = attrs
        next_url = acs_handler(mock_saml_auth, "/foo")

        assert next_url == "/foo"
        assert current_user.is_authenticated
        assert current_user.confirmed_at is None


def test_acs_handler_factory_config(appctx, db):
    """Test ACS handler factory with config."""

    appctx.config["SSO_SAML_IDPS"] = {
        "test": {
            "mappings": {
                "email": "email",
                "name": "name",
                "surname": "surname",
                "external_id": "external_id",
            },
            "auto_confirm": True,
        }
    }
    attrs = dict(
        email=["federico@example.com"],
        name=["federico"],
        surname=["Fernandez"],
        external_id=["12345679abcdf"],
    )

    acs_handler = acs_handler_factory("test")

    with appctx.test_request_context(), patch(
        "invenio_saml.utils.SAMLAuth"
    ) as mock_saml_auth:
        mock_saml_auth.get_attributes.return_value = attrs
        acs_handler(mock_saml_auth, "/foo")

        assert current_user.is_authenticated
        assert current_user.confirmed_at


def test_acs_handler_authetication_error(appctx, db):
    """Test ACS handler factory authentication errors."""
    attrs = dict(
        email=["federico@example.com"],
        name=["federico"],
        surname=["Fernandez"],
        external_id=["12345679abcdf"],
    )

    acs_handler = acs_handler_factory("test")

    with appctx.test_request_context(), patch(
        "invenio_saml.utils.SAMLAuth"
    ) as mock_saml_auth, patch(
        "invenio_saml.handlers.account_authenticate"
    ) as mock_authenticate:
        mock_saml_auth.get_attributes.return_value = attrs
        mock_authenticate.return_value = False
        with pytest.raises(Unauthorized):
            acs_handler(mock_saml_auth, "/foo")


def test_acs_handler_user_creation_error(appctx, db):
    """Test ACS handler factory user creation errors."""
    attrs = dict(
        email=["federico@example.com"],
        name=["federico"],
        surname=["Fernandez"],
        external_id=["12345679abcdf"],
    )

    acs_handler = acs_handler_factory("test")

    with appctx.test_request_context(), patch(
        "invenio_saml.utils.SAMLAuth"
    ) as mock_saml_auth, patch(
        "invenio_saml.handlers.account_register"
    ) as mock_register:
        mock_saml_auth.get_attributes.return_value = attrs
        mock_register.return_value = None
        with pytest.raises(Unauthorized):
            acs_handler(mock_saml_auth, "/foo")


def test_custom_account_info(appctx, db):
    """Test custom account info in ACS handler."""
    appctx.config["SSO_SAML_IDPS"] = {
        "test": {
            "auto_confirm": True,
        }
    }

    def account_info(attributes, remote_app):
        remote_app_config = current_app.config["SSO_SAML_IDPS"].get(remote_app, {})
        return dict(
            user=dict(
                email=attributes["email"],
                profile=dict(
                    username=attributes["name"],
                    full_name=f"{attributes['name']} {attributes['surname']}",
                ),
            ),
            external_id=f"{attributes['name']}.{attributes['surname']}",
            external_method=remote_app,
            active=True,
            confirmed_at=(
                datetime.now(timezone.utc)
                if remote_app_config.get("auto_confirm", False)
                else None
            ),
        )

    attrs = dict(
        email="federico@example.com",
        name="federico",
        surname="fernandez",
    )

    acs_handler = acs_handler_factory("test", account_info=account_info)

    with appctx.test_request_context(), patch(
        "invenio_saml.utils.SAMLAuth"
    ) as mock_saml_auth:
        mock_saml_auth.get_attributes.return_value = attrs
        acs_handler(mock_saml_auth, "/")

        assert current_user.is_authenticated
        assert current_user.confirmed_at


def test_custom_user_lookup(appctx, users):
    """Test custom user lookup in ACS handler."""
    appctx.config["SSO_SAML_IDPS"] = {
        "test": {
            "mappings": {
                "email": "email",
                "name": "name",
                "surname": "surname",
                "external_id": "external_id",
            },
        }
    }
    attrs = dict(
        email=["federico@example.com"],
        name=["federico"],
        surname=["Fernandez"],
        external_id=["12345679abcdf"],
    )

    user = current_datastore.get_user_by_email("federico@example.com")

    mock_user_lookup = Mock(return_value=user)

    acs_handler = acs_handler_factory("test", user_lookup=mock_user_lookup)

    with appctx.test_request_context(), patch(
        "invenio_saml.utils.SAMLAuth"
    ) as mock_saml_auth:
        mock_saml_auth.get_attributes.return_value = attrs
        acs_handler(mock_saml_auth, "/")

        assert mock_user_lookup.call_count == 1
        assert current_user.is_authenticated
        assert current_user.confirmed_at is None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. Garcia Gabancho.
# Copyright (C) 2019-2021 CERN.
# Copyright (C) 2021 Graz University of Technology.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Temporary place for, IMHO, Invenio-Accounts code.

All this code has been adapted and copied from Invenio-Oauthclient.
"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019, 2022 Esteban J. Garcia Gabancho.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that provides SAML integration."""

import json
from collections.abc import Mapping
from functools import wraps

from flask import url_for
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
from werkzeug.utils import cached_property, import_string

from . import config
from .errors import IdentityProviderNotFound
from .utils import SAMLAuth, prepare_flask_request
from .views import create_blueprint


def _default_config(idp):
    """Default IdP configuration."""
    return dict(
        settings={
            "strict": True,
            "debug": True,
            "sp": {
                "entityId": url_for("sso_saml.metadata", idp=idp, _external=True),
                "assertionConsumerService": {
                    "url": url_for("sso_saml.acs", idp=idp, _external=True),
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",
                },
                "singleLogoutService": {
                    "url": url_for("sso_saml.sls", idp=idp, _external=True),
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
                },
                "NameIDFormat": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
                "x509cert": "",
                "privateKey": "",
            },
            "idp": {
                "entityId": None,
                "singleSignOnService": {
                    "url": None,
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
                },
                "singleLogoutService": {
                    "url": None,
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
                },
                "x509cert": None,
            },
            "security": {
                "authnRequestsSigned": False,
                "failOnAuthnContextMismatch": False,
                "logoutRequestSigned": False,
                "logoutResponseSigned": False,
                "metadataCacheDuration": None,
                "metadataValidUntil": None,
                "nameIdEncrypted": False,
                "requestedAuthnContext": True,
                "requestedAuthnContextComparison": "exact",
                "signMetadata": False,
                "signatureAlgorithm": "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
                "wantAssertionsEncrypted": False,
                "wantAssertionsSigned": False,
                "wantMessagesSigned": False,
                "wantNameId": True,
                "wantNameIdEncrypted": False,
                "digestAlgorithm": "http://www.w3.org/2001/04/xmlenc#sha256",
            },
        },
        settings_file_path=None,
        settings_url=None,
        sp_cert_file=None,
        sp_key_file=None,
        settings_handler=None,
        login_handler=None,
        acs_handler=None,
        logout_handler=None,
        sls_handler=None,
    )


def _cached_configuration(f):
    """Cache the IdP configuration for future use."""

    @wraps(f)
    def inner(self, idp, *args, **kwargs):
        if idp not in self._saml_config:
            try:
                self._saml_config[idp] = self._build_configuration(idp)
            except KeyError as exc:
                raise IdentityProviderNotFound() from exc
        return f(self, idp, *args, **kwargs)

    return inner


class _InvenioSSOSAMLState(object):
    """Invenio SSO SAML state object."""

    def __init__(self, app):
        """Initialize state."""
        self.app = app
        self._saml_config = {}

    @property
    def url_prefix(self):
        """URL prefix from config."""
        return self.app.config["SSO_SAML_DEFAULT_BLUEPRINT_PREFIX"]

    @property
    def metadata_url(self):
        """SSO metadata URL from config."""
        return self.app.config["SSO_SAML_DEFAULT_METADATA_ROUTE"]

    @property
    def sso_url(self):
        """SSO SSO URL from config."""
        return self.app.config["SSO_SAML_DEFAULT_SSO_ROUTE"]

    @property
    def acs_url(self):
        """SSO ACS URL from config."""
        return self.app.config["SSO_SAML_DEFAULT_ACS_ROUTE"]

    @property
    def slo_url(self):
        """SSO SLO URL from config."""
        return self.app.config["SSO_SAML_DEFAULT_SLO_ROUTE"]

    @property
    def sls_url(self):
        """SSO SLS URL from config."""
        return self.app.config["SSO_SAML_DEFAULT_SLS_ROUTE"]

    @cached_property
    def prepare_flask_request(self):
        """Function to prepare flask request for OneLogin."""
        prep_func = self.app.config["SSO_SAML_PREPARE_FLASK_REQUEST_FUNCTION"]
        if isinstance(prep_func, str):
            prep_func = import_string(prep_func)
        return prep_func

    @_cached_configuration
    def get_settings(self, idp):
        """Find settings for a particular Identity Provider."""
        return self._saml_config[idp]["settings"]

    @_cached_configuration
    def get_handler(self, idp, handler):
        """Get handler for idp."""
        return self._saml_config[idp][handler]

    def get_auth(self, idp):
        """Instantiate the IdP."""
        return SAMLAuth(idp, self.get_settings(idp))

    def _build_configuration(self, idp):
        """Update default config with the ones read from configuration."""

        def update(d, u):
            for k, v in u.items():
                if isinstance(v, Mapping):
                    d[k] = update(d.get(k, {}), v)
                else:
                    d[k] = v
            return d

        def make_handler(handler, default=None):
            handler = handler if handler else default
            return (
                import_string(handler)
                if handler and isinstance(handler, str)
                else handler
            )

        config = _default_config(idp)
        update(config, self.app.config["SSO_SAML_IDPS"][idp])

        # Read IdP config from file or URL if any
        if config["settings_url"]:
            external_conf = OneLogin_Saml2_IdPMetadataParser.parse_remote(
                config["settings_url"]
            )
            config["settings"]["idp"].update(external_conf.get("idp"))

        if config["settings_file_path"]:
            with open(config["settings_file_path"], "r") as idp:
                file = config["settings_file_path"]
                # xml format
                if file.endswith(".xml"):
                    external_conf = OneLogin_Saml2_IdPMetadataParser.parse(idp.read())
                # json format
                elif file.endswith(".json"):
                    external_conf = json.loads(idp.read())
            config["settings"]["idp"].update(external_conf.get("idp"))

        # Load certificate and key
        if config["sp_cert_file"]:
            with open(config["sp_cert_file"], "r") as cf:
                cert = cf.read()
            config["settings"]["sp"]["x509cert"] = cert

        if config["sp_key_file"]:
            with open(config["sp_key_file"], "r") as cf:
                cert = cf.read()
            config["settings"]["sp"]["privateKey"] = cert

        # Import handlers is present
        config["settings_handler"] = make_handler(
            config["settings_handler"],
            self.app.config.get("SSO_SAML_DEFAULT_SETTINGS_HANDLER"),
        )
        config["login_handler"] = make_handler(
            config["login_handler"],
            self.app.config.get("SSO_SAML_DEFAULT_LOGIN_HANDLER"),
        )
        config["logout_handler"] = make_handler(
            config["logout_handler"],
            self.app.config.get("SSO_SAML_DEFAULT_LOGOUT_HANDLER"),
        )
        config["acs_handler"] = make_handler(
            config["acs_handler"],
            self.app.config.get("SSO_SAML_DEFAULT_ACS_HANDLER"),
        )
        config["sls_handler"] = make_handler(
            config["sls_handler"],
            self.app.config.get("SSO_SAML_DEFAULT_SLS_HANDLER"),
        )

        return config


class InvenioSSOSAML(object):
    """Invenio-SSO-SAML extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)

        state = _InvenioSSOSAMLState(app)

        # Register blueprint and routes
        app.register_blueprint(create_blueprint(state, __name__))

        app.extensions["invenio-sso-saml"] = state
        return state

    def init_config(self, app):
        """Initialize configuration."""
        # Only load the configuration, we'll build it lazily because we can
        for k in dir(config):
            if k.startswith("SSO_SAML_"):
                app.config.setdefault(k, getattr(config, k))

        # Set default when needed
        app.config.setdefault(
            "SSO_SAML_PREPARE_FLASK_REQUEST_FUNCTION", prepare_flask_request
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2024 Esteban J. Garcia Gabancho.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test code that will get externalized at some point."""

import pytest

from invenio_saml.invenio_app import get_safe_redirect_target


@pytest.mark.parametrize(
    "next_url,expected",
    [
        ("/", "/"),
        ("/foo", "/foo"),
        ("http://not-safe.com", None),
        ("https://example.com", "https://example.com"),
    ],
)
def test_get_safe_redirect(base_app, next_url, expected):
    """Test get safe redirect target."""
    with base_app.test_request_context(query_string={"next": next_url}):
        safe_next = get_safe_redirect_target()
        assert safe_next == expected



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2024 Esteban J. Garcia Gabancho.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module that provides SAML integration.

This module provides you with a set of default handlers than can be used out of
the box to authenticated users using SSO SAML.

Handlers
--------

Invenio-SAML allows you to specify handlers for each SSO action, ``sso``,
``acs``, ``slo``, ``sls`` and ``metadata``.
Typically the ones that matter the most are ``acs`` and ``sls``, because they
are the ones that will authenticate and "unauthenticate" users from the
application.

This module provides default handlers for ``acs`` and``sls`` actions that will
be valid for most of the use cases. The default ACS handler is created by a factory,
:func: `invenio_saml.handlers.acs_handler_function`.
This is how you can use them:

.. code-block:: python

    from invenio_saml.handlers import acs_handler_factor

    SSO_SAML_IDPS={
            '<idp-name>': {
                'settings': {
                    'idp': {
                        'entityId': '<idp-url>',
                        'singleSignOnService': {
                            'url': '<ipd-sso-url>',
                        },
                        'singleLogoutService': {
                            'url': '<idp-slo-url>',
                        },
                        'x509cert': '<ipd-cert>',
                    },
                },
                "mappings": {
                    "email": "User.email",
                    "name": "User.FirstName",
                    "surname": "User.LastName",
                    "external_id": "User.email",
                },
                'acs_handler': acs_handler_factory('<idp-name>'),
                'auto_confirm': False,
            }
        }
"""

from .ext import InvenioSSOSAML

__version__ = "1.1.0"

__all__ = ("__version__", "InvenioSSOSAML")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. Garcia Gabancho.
# Copyright (C) 2019-2021 CERN.
# Copyright (C) 2021 Graz University of Technology.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Temporary file for, IMHO, Invenio-App code.

All this code has been adapted and copied from Invenio-Oauthclient.
"""

from flask import current_app, request
from uritools import uricompose, urisplit


def get_safe_redirect_target(arg="next", _target=None):
    """Get URL to redirect to and ensure that it is local.

    :param arg: URL argument.
    :returns: The redirect target or ``None``.
    """
    for target in _target, request.args.get(arg), request.referrer:
        if target:
            redirect_uri = urisplit(target)
            allowed_hosts = current_app.config.get("APP_ALLOWED_HOSTS", [])
            if redirect_uri.host in allowed_hosts:
                return target
            elif redirect_uri.path:
                return uricompose(
                    path=redirect_uri.path,
                    query=redirect_uri.query,
                    fragment=redirect_uri.fragment,
                )
    return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Esteban J. Garcia Gabancho.
# Copyright (C) 2019-2021 CERN.
# Copyright (C) 2021-2022 Graz University of Technology.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Temporary file for, IMHO, Invenio-Accounts code."""

from __future__ import absolute_import, print_function

from flask import after_this_request, current_app
from flask_security import login_user
from flask_security.confirmable import requires_confirmation
from flask_security.registerable import register_user

# FIXME: modify import when integrated inside invenio_accounts
# from .models import User
from invenio_accounts.models import User
from invenio_oauthclient.models import UserIdentity
from werkzeug.local import LocalProxy

_security = LocalProxy(lambda: current_app.extensions["security"])

_datastore = LocalProxy(lambda: _security.datastore)


def _commit(response=None):
    _datastore.commit()
    return response


def _get_external_id(account_info):
    """Get external id from account info."""
    if all(k in account_info for k in ("external_id", "external_method")):
        return dict(
            id=account_info["external_id"], method=account_info["external_method"]
        )
    return None


def account_get_user(account_info=None):
    """Retrieve user object for the given request.

    Uses either the access token or extracted account information to retrieve
    the user object.

    :param account_info: The dictionary with the account info.
        (Default: ``None``)
    :returns: A :class:`invenio_accounts.models.User` instance or ``None``.
    """
    if account_info:
        external_id = _get_external_id(account_info)
        if external_id:
            user = UserIdentity.get_user(external_id["method"], external_id["id"])
            if user:
                return user

        email = account_info.get("user", {}).get("email")
        if email:
            return User.query.filter_by(email=email).one_or_none()
    return None


def account_authenticate(user):
    """Authenticate an ACS callback.

    :param user: A user instance.
    :returns: ``True`` if the user is successfully authenticated.
    """
    if not requires_confirmation(user):
        after_this_request(_commit)
        return login_user(user, remember=False)
    return False


def account_link_external_id(user, external_id=None):
    """Link a user to an external id.

    :param user: A :class:`invenio_accounts.models.User` instance.
    :param external_id: The external id associated with the user.
        (Default: ``None``)
    :raises invenio_oauthclient.errors.AlreadyLinkedError: Raised if already
        exists a link.
    """
    if UserIdentity.get_user(external_id["method"], external_id["id"]):
        # already linked. should be fine to just return
        # method and id form the composite primary key, so no other with these values can be linked
        return

    UserIdentity.create(user, external_id["method"], external_id["id"])


def account_register(form, confirmed_at=None):
    """Register user if possible.

    :param form: A form instance.
    :returns: A :class:`invenio_accounts.models.User` instance.
    """
    if form.validate():
        data = {
            **form.to_dict(),
            "confirmed_at": confirmed_at,
        }
        if not data.get("password"):
            data["password"] = ""
        user = register_user(**data)
        if not data["password"]:
            user.password = None
        _datastore.commit()
        return user



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2024 Esteban J. Garcia Gabancho.
# Copyright (C) 2021-2022 Graz University of Technology.
#
# Invenio-SAML is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Default handlers for SSO-SAML."""

from datetime import datetime, timezone

from flask import abort, current_app
from flask_login import current_user
from flask_security import logout_user
from invenio_db import db
from invenio_oauthclient.errors import AlreadyLinkedError
from invenio_oauthclient.utils import create_csrf_disabled_registrationform, fill_form

from .invenio_accounts.utils import (
    account_authenticate,
    account_get_user,
    account_link_external_id,
    account_register,
)
from .invenio_app import get_safe_redirect_target


def default_account_info(attributes, remote_app):
    """Return account info for remote user.

    This function uses the mappings configuration variable inside your IdP
    configuration.

    :param attributes: (dict) dictionary of data returned by identity provider.
    :param remote_app: (str) Identity provider key.

    :returns: (dict) A dictionary representing user to create or update.
    """
    remote_app_config = current_app.config["SSO_SAML_IDPS"][remote_app]

    mappings = remote_app_config["mappings"]

    affiliations_mapping = mappings.get("affiliations", None)
    if affiliations_mapping:
        affiliations = attributes[affiliations_mapping]
    else:
        affiliations = ""

    name = attributes[mappings["name"]][0]
    surname = attributes[mappings["surname"]][0]
    email = attributes[mappings["email"]][0]
    external_id = attributes[mappings["external_id"]][0]
    username = (
        remote_app + "-" + external_id.split("@")[0]
        if "@" in external_id
        else remote_app + "-" + external_id
    )

    return dict(
        user=dict(
            email=email,
            profile=dict(
                username=username,
                full_name=name + " " + surname,
                affiliations=affiliations,
            ),
        ),
        external_id=external_id,
        external_method=remote_app,
        active=True,
        confirmed_at=(
            datetime.now(timezone.utc)
            if remote_app_config.get("auto_confirm", False)
            else None
        ),
    )


def default_account_setup(user, account_info):
    """Default account setup which only links ``User`` and ``UserIdentity``."""
    try:
        account_link_external_id(
            user,
            dict(
                id=account_info["external_id"], method=account_info["external_method"]
            ),
        )
    except AlreadyLinkedError:
        pass


def default_sls_handler(auth, next_url):
    """Default SLS handler which simply logs out the user."""
    logout_user()
    next_url = (
        get_safe_redirect_target(_target=next_url)
        or current_app.config["SECURITY_POST_LOGOUT_VIEW"]
    )
    return next_url


def acs_handler_factory(
    remote_app,
    account_info=default_account_info,
    account_setup=default_account_setup,
    user_lookup=account_get_user,
):
    """Generate ACS handlers with an specific account info and setup functions.

    .. note::

        In 90% of the cases the ACS handler is going to be the same, only the
        way the information is extracted and processed from the IdP will be
        different.

    :param remote_app: string representing the name of the identity provider.

    :param account_info: callable to extract the account information from a
        dict like object. ``mappings`` key is required whe using it.
        This function is expected to return a dictionary similar to this:

        .. code-block:: python

            dict(
                user=dict(
                    email='federico@example.com',
                    profile=dict(username='federico',
                                 full_name='Federico Fernandez'),
                ),
                external_id='12345679abcdf',
                external_method='example',
                active=True
             )

        Where ``external_id`` is the ID provided by the IdP and
        ``external_method`` is the name of the IdP as in the configuration
        file (not mandatory but recommended).

    :param account_setup: callable to setup the user account with the
        corresponding IdP account information. Typically this means creating a
        new row under ``UserIdentity`` and maybe extending  ``g.identity``.

    :param user_lookup: callable to retrieve any user whose information matches
        what is returned by the `account_info` callable. This then returns a
        User object if a match is present and None if no match is found.

    :return: function to be used as ACS handler
    """

    def default_acs_handler(auth, next_url):
        """Default ACS handler.

        :para auth: A :class:`invenio_saml.utils.SAMLAuth` instance.
        :param next_url: String with the next URL to redirect to.

        :return: Next URL
        """
        if not current_user.is_authenticated:
            current_app.logger.debug(
                "Metadata received from IdP %s", auth.get_attributes()
            )
            _account_info = account_info(auth.get_attributes(), remote_app)
            current_app.logger.debug("Metadata extracted from IdP %s", _account_info)
            # TODO: signals?

            user = user_lookup(_account_info)

            if user is None:
                form = create_csrf_disabled_registrationform(remote_app)
                form = fill_form(form, _account_info["user"])
                user = account_register(
                    form, confirmed_at=_account_info["confirmed_at"]
                )

            # if registration fails ... TODO: signup?
            if user is None or not account_authenticate(user):
                abort(401)

            account_setup(user, _account_info)

        db.session.commit()

        next_url = (
            get_safe_redirect_target(_target=next_url)
            or current_app.config["SECURITY_POST_LOGIN_VIEW"]
        )
        return next_url

    return default_acs_handler
